#include <time.h>
#include <sys/time.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <unistd.h>
/* the following code sets up a signal hander (sigaction) and fires off
 * signals ( using either itimer or posix4 methods if given 0 or 1 as 
 * the command line option respectively ).
 */
#define TIME_MSECS 1500
void handler_itimer(
                    int sig,
                    siginfo_t *info,
                    void *context
                    );
void handler_posix(
                   int sig,
                   siginfo_t *info,
                   void *context
                   );
int main(
         int argc,
         char **argv
         )
{
  int
    timer; /* 0 itimer, 1 posix */
  int
    time_secs, time_msecs; /* parts of time */
  int
    signal; /* the signal used */
  /* ------------ */
  /* itimer stuff */
  /* ------------ */
  int which_itimer; /* sent to setitimer */
  struct itimerval
    value,        /* new value */
    value_old;    /* old value */
  /* ------------------ */
  /* POSIX4 timer stuff */
  /* ------------------ */
  timer_t
    psxtimer;    /* POSIX4 timer */
  struct sigevent
    psxsigevent; /* POSIX4 signal event */
  struct itimerspec
    psxspec,     /* new value */
    psxspec_old; /* old value */
  /* ------------ */
  /* signal stuff */
  /* ------------ */
  struct sigaction
    action,         /* new action */
    action_old;     /* old action */
  /* command line args */
  if( argc != 2 )
    {
      printf( "Usage: signal_handler <0|1>\n" );
      printf( "  0: itimer\n" );
      printf( "  1: posix\n" );
   return( 1 );
    }
  if( strcmp( argv[1], "0" ) == 0 )
    {
      timer  = 0;
      signal = SIGALRM;
    }
  else
    {
      timer  = 1;
      signal = SIGRTMAX;
    }
  time_secs  = TIME_MSECS/1000;
  time_msecs = TIME_MSECS - time_secs*1000;
  /* -------------------------- */
  /* sigaction - signal handler */
  /* -------------------------- */
  memset(&(action_old),0x00,sizeof(struct sigaction));
  memset(&(action),0x00,sizeof(struct sigaction));
  action.sa_flags = SA_RESTART|SA_SIGINFO;
  if( timer == 0 )
    {
      action.sa_sigaction = handler_itimer;
    }
  else
    {
      action.sa_sigaction = handler_posix;
    }
  if (sigaction( signal, &(action), &(action_old)) != 0)
    {
      printf( "bad sigaction\n" );
      return( 1 );
    }
  /* ------------------------------------------ */
  /* itimer                                     */
  /* Set timer to send signal at time intervals */
  /* ------------------------------------------ */
  if( timer == 0 )
    {
      printf( "itimer\n" );
      signal                    = SIGALRM;
      which_itimer              = ITIMER_REAL;
      value.it_interval.tv_sec  = time_secs;
      value.it_interval.tv_usec = time_msecs*1000;
      value.it_value.tv_sec     = time_secs;
      value.it_value.tv_usec    = time_msecs*1000;
      if (setitimer(which_itimer, &(value),
                    &(value_old)) != 0)
        {
          printf( "bad setitimer\n" );
          return( 1 );
        }
    }
  /* ------------------------------------------ */
  /* POSIX4                                     */
  /* Set timer to send signal at time intervals */
  /* ------------------------------------------ */
  else
    {
      printf( "posix\n" );
      signal                            = SIGRTMAX;
      psxsigevent.sigev_notify          = SIGEV_SIGNAL;
      psxsigevent.sigev_signo           = signal;
      psxsigevent.sigev_value.sival_int = 13;
      psxspec.it_interval.tv_nsec       = time_msecs*1000000;
      psxspec.it_interval.tv_sec        = time_secs;
      psxspec.it_value.tv_nsec          = time_msecs*1000000;
      psxspec.it_value.tv_sec           = time_secs;
      if( timer_create( CLOCK_REALTIME, &(psxsigevent), &(psxtimer) ) != 0 )
        {
          printf( "bad timer_create\n" );
          return( -1 );
        }
      if( timer_settime( psxtimer, 0, &(psxspec), &(psxspec_old) ) != 0 )
        {
          printf( "bad timer_settime\n" );
          return( -1 );
        }
    }
  printf( "Signals every %d msecs (ctrl-C to exit)\n", TIME_MSECS );
  while(1){}
  return( 0 );
}
void handler_itimer(
                    int sig,
                    siginfo_t *info,
                    void *context
                    )
{
  static int i=0;
  printf( "handler_itimer %d (ctrl-C to exit)\n", i++ );
  if( i > 10 )
    abort();
}
void handler_posix(
                   int sig,
                   siginfo_t *info,
                   void *context
                   )
{
  static int i=0;
  printf( "handler_posix %d (ctrl-C to exit)\n", i++ );
  if( i > 10 )
    abort();
}
