eval 'exec perl -w -S $0 ${1+"$@"}'
  if 0;

if ( $#ARGV < 0 || $ARGV[0] =~ /^--?help/)
  {
    print <<'EOF';
#............................................................................
#...Name
#...====
#... procmon_post.pl
#...
#...Usage
#...=====
#... procmon_post.pl
#...   [-filter <memory>]
#...   <procmon output file list>
#...
#... Options
#... -------
#...  -filter <memory>
#...   Filter out any code blocks in which there was no memory
#...   gain.
#...
#...Purpose
#...=======
#... Takes the output files of procmon sampling and does some simple analysis
#... and creates plots (procmon.ps and procmon.pdf).
#...
#... Gnuplot data files are produced (1 per input data file).
#... The format for these files is:
#...       <procmon output file name>.dat
#...
#... This script uses gnuplot and ps2pdf (they must be in the PATH).
#...
#...Examples
#...========
#... 1) Plot data from one file (ie one process):
#...      procmon_post.pl procmon_qsc20_11131568.txt
#... 2) Plot data from many files (ie all process from the run)
#...    and filter out code blocks where there was no memory gain.
#...      procmon_post.pl procmon_*.txt -filter memory
#............................................................................
EOF
  exit;
  }
#............................................................................
#...Program Flow
#...============
#...1) get earliest time
#...2) process each file
#...3) finish statting results_all
#...4) print results
#...5) plot
#............................................................................
#....................................................
#...global field names and list of required fields...
#....................................................
$G_FN_EPOCH_TIME = "EpochTime";
$G_FN_PWALLTIME  = "pWallTime";
$G_FN_PSIZE      = "pSize";
$G_FN_MMEMSIZE   = "MMEMSIZE";
$G_FN_TAG        = "TAG";
$G_FN_COMMAND    = "COMMAND";
@G_FNREQ = (
            $G_FN_EPOCH_TIME,
            $G_FN_PWALLTIME,
            $G_FN_PSIZE,
            $G_FN_MMEMSIZE,
            $G_FN_COMMAND
           );
#.........................
#...special block names...
#.........................
#...have space in name to distinguish from user defined...
$G_B_FINAL_RES = "Final Res";
$G_B_ORIG_DATA = "Orig Data";
#...........................................
#...If warnings have already been printed...
#...........................................
$G_WARN_P_START = 0;
$G_WARN_P_STOP = 0;
#...................
#...valid filters...
#...................
$G_CMD_FILTER_VALID = "memory"; # regexp of valid filters
#....................
#...cmd hash names...
#....................
$G_CMD_FILENAMES = "filenames";
$G_CMD_FILTER    = "filter";
#.............
#...my vars...
#.............
my(
   $basetime, # earliest time (for translating epoch times relative time)
   $block, # current block
   %cmd, # command line options
   @data, # list of data and field values for each sampling
   %field_info, # info about fields (definition and order)
   $filenum, # current file number
   $ierr, # error return value
   $gnuplot_info, # info gnuplot will need to plot
   $maxtime, # the maximum time
   %results, # results for a file
   %results_all, # all results from all files
   %start_stop, # starting/stopping idx in data for each instance of blocks
   %trace_data, # $trace_data{stuff} = val, {child}{<block>} = child struct
   %trace_data_all, # $trace_data_all{stuff} = val, {child}{<block>} = child
  );
#..........
#...init...
#..........
&parse_command_line( \@ARGV, \%cmd );
$field_info{block}{def}{delta_t} = "Time in Block (secs)";
$field_info{block}{def}{max_m} = "Maximum Memory Used (bytes)";
$field_info{block}{def}{delta_m} = "Change in Memory Size (bytes)";
@{$field_info{block}{order}} = (
                                "delta_t",
                                "delta_m",
                                "max_m"
                               );
#.......................
#...get earliest time...
#.......................
&get_basetime( \@{$cmd{$G_CMD_FILENAMES}}, \$basetime );
#.......................
#...process each file...
#.......................
undef( %results_all );
undef( %gnuplot_info );
undef( %trace_data_all );
$filenum = 0;
$maxtime = 0;
foreach $filename ( @{$cmd{$G_CMD_FILENAMES}} )
  {
    $filenum++;
    &parse_file( $filename, $basetime, \%start_stop, \%field_info,
                 \@data, \%trace_data, \$maxtime );
    &get_results( \%start_stop, \@data, $filenum, \%results,
                  \%results_all );
    $ierr = &print_data_info( \@data, \%field_info, $filenum, $filename,
                              \%gnuplot_info );
    if( $ierr != 0 )
      {
        &print_error( "Error return from print_data_info",
                      $ierr );
        exit( $ierr );
      }
    &get_results_trace( \%trace_data, \@data, $filenum, \%trace_data_all );
  }
#....................................
#...finish statting trace_data_all...
#....................................
&trace_data_all_finish( \%trace_data_all, \%trace_data_all );
#...................
#...print results...
#...................
&print_results_all( \%results_all, \%cmd );
&print_trace_data_all( \%trace_data_all, \%cmd );
#..........
#...plot...
#..........
&plot( \%gnuplot_info, \%field_info, \@{$cmd{$G_CMD_FILENAMES}},
       $maxtime, \%results_all, \%cmd );
#........................
#...log software usage...
#........................
&log_lanl_usage( "procmon", $#{$cmd{$G_CMD_FILENAMES}}+1 );
#.........
#...end...
#.........
exit( 0 );
#............................................................................
#...Name
#...====
#... parse_command_line
#...
#...Purpose
#...=======
#... Parses the command line arguments.
#...
#...Arguments
#...=========
#... argv_ref          Intent: in
#...                   Perl type: reference to array
#...                   reference to @ARGV
#...
#... cmd_ref           Intent: out
#...                   Perl type: reference to hash
#...                   reference %cmd hash
#...
#...Program Flow
#...============
#... 1) open every file
#... 2) read until you get to the first epoch time
#... 3) keep track of minimum one.
#............................................................................
sub parse_command_line
  {
    my(
       $argv_ref,
       $cmd_ref,
       ) = @_;
    my(
       $i, # loop var
       $opt, # current option
       $val, # value to current option
      );
    $i = 0;
    while( $i <= $#$argv_ref )
      {
        $opt = $$argv_ref[$i];
        if( $opt =~ /-filter/ )
          {
            if( $i == $#$argv_ref )
              {
                $ierr = 1;
                &print_error( "Missing argument to option [$opt]",
                              $ierr );
                exit( $ierr );
              }
            else
              {
                $i++;
                $val = $$argv_ref[$i];
                if( $val !~ /^$G_CMD_FILTER_VALID$/ )
                  {
                    $ierr = 1;
                    &print_error( "Incorrect value [$val] for option [$opt]",
                                  "Valid values = [$G_CMD_FILTER_VALID]",
                                  $ierr );
                    exit( $ierr );
                  }
                $$cmd_ref{$G_CMD_FILTER}{$val} = "";
                $i++;
              }
          }
        else
          {
            push( @{$$cmd_ref{$G_CMD_FILENAMES}}, $$argv_ref[$i] );
            $i++;
          }
      }
  }
#............................................................................
#...Name
#...====
#... get_basetime
#...
#...Purpose
#...=======
#... Gets the earliest EpochTime from the files to shift time to start
#... at 0.
#...
#...Arguments
#...=========
#... filenames_ref     Intent: in
#...                   Perl type: reference to array
#...                   reference to array of file names.
#...
#... basetime_ref      Intent: out
#...                   Perl type: reference to scalar
#...                   reference to basetime scalar.
#...
#...Program Flow
#...============
#... 1) open every file
#... 2) read until you get to the first epoch time
#... 3) keep track of minimum one.
#............................................................................
sub get_basetime
  {
    my(
       $filenames_ref,
       $basetime_ref
       ) = @_;
    my(
       $filename, # a filename
       $ierr, # error return value
       $line, # single line of a file
       $time, # time for this file
      );
    $$basetime_ref = 1e99;
    foreach $filename ( @$filenames_ref )
      {
        if( ! open( FILE, $filename ) )
          {
            $ierr = 1;
            &print_error( "Cannot open file [$filename]", $ierr );
            exit( $ierr );
          }
        while( $line=<FILE> )
          {
            if( $line =~ /^PROCMON\s+(\S+)/ )
              {
                $time = $1;
                if( $time < $$basetime_ref )
                  {
                    $$basetime_ref = $time;
                  }
                last;
              }
          }
        close( FILE );
      }
  }
#............................................................................
#...Name
#...====
#... parse_file
#...
#...Purpose
#...=======
#... Gets file data for a file.
#...
#...Arguments
#...=========
#... filename          Intent: in
#...                   Perl type: scalar
#...                   The name of the file
#...
#... basetime          Intent: in
#...                   Perl type: scalar
#...                   The basetime to adjust time to.
#...
#... start_stop_ref    Intent: out
#...                   Perl type: reference to hash
#...                   Start/stop data indices per instance of block
#...
#... field_info_ref    Intent: inout
#...                   Perl type: reference to hash
#...                   Info about fields (descriptions and order)
#...
#... data_ref          Intent: out
#...                   Perl type: reference to hash
#...                   For each line, contains all field data
#...
#... trace_data_ref    Intent: out
#...                   Perl type: reference to hash
#...                   Will flesh out:
#...                      $trace_data{block}{start_stop}[]:
#...                        array of start/stop line numbers
#...                      $trace_data{block}{child}{block_child}:
#...                        child block (points to same struct)
#...
#... maxtime_ref       Intent: inout
#...                   Perl type: reference to scalar
#...                   The new maximum time (if greater than the original value)
#...
#...Program Flow
#...============
#... 1) foreach line of file
#... 1.1) if field info, stuff field_info
#... 1.2) if fields, stuff into @fields for use when stuffing data
#... 1.3) if data
#... 1.3.1) get starting line
#... 1.3.2) if data line
#... 1.3.2.1) stuff each field into data
#... 1.3.2.2) pust TAG into start_stop
#... 1.3.3) adjust EpochTime with basetime
#... 2) add final result block (first and last data line)
#............................................................................
sub parse_file
  {
    my(
       $filename,
       $basetime,
       $start_stop_ref,
       $field_info_ref,
       $data_ref,
       $trace_data_ref,
       $maxtime_ref
       ) = @_;
    my(
       $block, # (val):(block) in TAG
       $block_point, # block name when constructing trace
       $data_idx, # data index
       $field, # the field name
       @fields, # the fields in this file
       @fields_new, # new values for fields
       $field_desc, # the field description
       $found, # boolean for finding something
       $i, # loop var
       $j, # loop var
       $ierr, # error return value
       $line, # the current line of the file
       $linec, # a copy of the current line
       $linenum, # the actual linenumber in filename
       $seen_cutoff, # if seen the field (MMEMSIZE) that you ignore onward
       $startline, # for the final results, the start line in the original file
       @trace_arr, # current block trace
       @trace_pop, # pop of trace - trace of correct block being popped
       $traces_ref, # reference in %traces hash
       %traces, # $tracee{$block}[] = array of traces for a block
       $val, # (val):(block) in TAG
      );
    if( ! open( FILE, $filename ) )
      {
        $ierr = 1;
        &print_error( "Cannot open file [$filename]", $ierr );
        exit( $ierr );
      }
    $data_idx = 0;
    undef( @$data_ref );
    undef( %$start_stop_ref );
    undef( %$trace_data_ref );
    $linenum = 0;
    $startline = -1;
    $seen_cutoff = 0;
    undef( @trace_arr );
    #.......................
    #...process each line...
    #.......................
    while( $line=<FILE> )
      {
        $linenum++;
        #................
        #...field info...
        #................
        if( $line =~ /^\#PROCMON:key\s+(\S+) = (.*)/ )
          {
            $field = $1;
            $field_desc = $2;
            if( !defined( $$field_info_ref{"$G_B_ORIG_DATA"}{def}{$field} ) )
              {
                $$field_info_ref{"$G_B_ORIG_DATA"}{def}{$field} = $field_desc;
                #.........................................................
                #...get new order baed on location relative to MMEMSIZE...
                #.........................................................
                undef( @fields );
                if( defined($$field_info_ref{"$G_B_ORIG_DATA"}{order}) )
                  {
                    @fields = @{$$field_info_ref{"$G_B_ORIG_DATA"}{order}};
                  }
                undef( @fields_new );
                #........................................
                #...stuff up to MMEMSIZE on fields_new...
                #........................................
                while( $#fields >= 0 && $fields[0] ne "$G_FN_MMEMSIZE" )
                  {
                    push( @fields_new, shift( @fields ) );
                  }
                #...........................................................
                #...if passed cutoff, stuff through command on fields_new...
                #...........................................................
                if( $seen_cutoff == 1 && $#fields >= 0 )
                  {
                    while( $#fields >= 0 && $fields[0] ne "$G_FN_COMMAND" )
                      {
                        push( @fields_new, shift( @fields ) );
                      }
                  }
                #...................................
                #...push this field on fields_new...
                #...................................
                push( @fields_new, $field );
                #.........................................
                #...push remaining fields on fields_new...
                #.........................................
                if( $#fields >= 0 )
                  {
                    push( @fields_new, @fields );
                  }
                #...update seen_cutoff last to account for adding mmemsize...
                #...itself...
                if( $field eq $G_FN_MMEMSIZE )
                  {
                    $seen_cutoff = 1;
                  }
                @{$$field_info_ref{"$G_B_ORIG_DATA"}{order}} = @fields_new;
              }
          }
        #............
        #...fields...
        #............
        elsif( $line =~ /^\#PROCMON / )
          {
            @fields = split( /\s+/, $line );
            shift( @fields );
          }
        #..........
        #...data...
        #..........
        elsif( $line =~ /^PROCMON\s+(.*)/)
          {
            #..............................................
            #...get the starting line of first data line...
            #..............................................
            if( $startline < 0 )
              {
                $startline = $linenum;
              }
            #..............................................
            #...strip of fields and stuff them into data...
            #...fields with same name are overwritten   ...
            #..............................................
            $linec = $1;
            foreach $field ( @fields )
              {
                if( $field eq "$G_FN_COMMAND" )
                  {
                    $$data_ref[$data_idx]{$field} = $linec;
                  }
                elsif( $linec =~ /\s*(\S+)\s+(.*)/ )
                  {
                    $$data_ref[$data_idx]{$field} = $1;
                    $linec = $2;
                  }
              }
            #...............................
            #...check for TAG: start/stop...
            #...............................
            if( defined( $$data_ref[$data_idx]{$G_FN_TAG} ) &&
                $$data_ref[$data_idx]{$G_FN_TAG} =~ /^(\d):(.*)/ )
              {
                $val = $1;
                $block = $2;
                #........................................
                #...push starting onto start_stop list...
                #........................................
                if( $val == 0 )
                  {
                    push( @{$$start_stop_ref{$block}}, "$data_idx:$linenum" );
                    push( @trace_arr, $block );
                    @{$traces{$block}[$#{$traces{$block}}+1]} = @trace_arr;
                    $traces_ref = $trace_data_ref;
                    foreach $block_point (@trace_arr)
                      {
                        $traces_ref = \%{$$traces_ref{child}{$block_point}};
                      }
                    push( @{$$traces_ref{start_stop}}, "$data_idx:$linenum" );
                  }
                #......................................................
                #...find last open start and close it with this stop...
                #......................................................
                else
                  {
                    #............................
                    #...remove from @trace_arr...
                    #............................
                    for( $i = $#trace_arr; $i >= 0; $i-- )
                      {
                        if( $trace_arr[$i] eq $block )
                          {
                            $trace_arr[$i] = "";
                            last;
                          }
                      }
                    #......................
                    #...trace start_stop...
                    #......................
                    if( defined $traces{$block} )
                      {
                        @trace_arr = grep( /\S/, @trace_arr );
                        @trace_pop = @{pop( @{$traces{$block}} )};
                        $traces_ref = $trace_data_ref;
                        foreach $block_point (@trace_pop)
                          {
                            $traces_ref = \%{$$traces_ref{child}{$block_point}};
                          }
                        for( $i = $#{$$traces_ref{start_stop}}; $i >= 0; $i-- )
                          {
                            if( $$traces_ref{start_stop}[$i] !~ / / )
                              {
                                $$traces_ref{start_stop}[$i] .=
                                  " $data_idx:$linenum";
                                last;
                              }
                          }
                      }
                    #...................................................
                    #...need this protecting if-defined because if no...
                    #...elements in array, will create a faux one    ...
                    #...................................................
                    $found = 0;
                    if( defined $$start_stop_ref{$block} )
                      {
                        for( $i = $#{$$start_stop_ref{$block}}; $i >= 0; $i-- )
                          {
                            if( $$start_stop_ref{$block}[$i] !~ / / )
                              {
                                $$start_stop_ref{$block}[$i] .= " $data_idx:$linenum";
                                $found = 1;
                                last;
                              }
                          }
                      }
                    if( $found == 0 )
                      {
                        #....................................................
                        #...nothing needed to be done except print warning...
                        #....................................................
                        if( ! $G_WARN_P_START )
                          {
                            $G_WARN_P_START = 1;
                            &print_error( "Corresponding 'start' not found ".
                                          "for block [$block]",
                                          "Ignoring this occurrance of the block.",
                                          "Ignoring future warnings of this ".
                                          "type",
                                          "File: $filename:$linenum",
                                          0 );
                          }
                      }
                  }
                #............................................................
                #...DONE: find last open start and close it with this stop...
                #............................................................
              }
            #.....................................
            #...DONE: check for TAG: start/stop...
            #.....................................
            #.................
            #...adjust time...
            #.................
            $$data_ref[$data_idx]{$G_FN_EPOCH_TIME} -= $basetime;
            $data_idx++;
          }
        #................
        #...DONE: data...
        #................
      }
    #.............................
    #...DONE: process each line...
    #.............................
    #....................
    #...update maxtime...
    #....................
    if( $#$data_ref >= 0 && $$data_ref[-1]{$G_FN_EPOCH_TIME} > $$maxtime_ref )
      {
        $$maxtime_ref = $$data_ref[-1]{$G_FN_EPOCH_TIME};
      }
    #...........................................
    #...add in first and last line as a block...
    #...........................................
    push( @{$$trace_data_ref{start_stop}},
          "0:$startline $#$data_ref\:$linenum" );
    push( @{$$start_stop_ref{"$G_B_FINAL_RES"}},
          "0:$startline $#$data_ref\:$linenum" );
    #......................................
    #...check for start_stop consistency...
    #......................................
    foreach $block ( sort keys %$start_stop_ref )
      {
        for( $i = 0; $i <= $#{$$start_stop_ref{$block}}; $i++ )
          {
            ($start, $stop) = split( / /, $$start_stop_ref{$block}[$i] );
            #.............................................................
            #...If stop not defined, just take last reading as the stop...
            #...Happens if program was terminated inside block         ...
            #.............................................................
            if( ! defined $stop )
              {
                $$start_stop_ref{$block}[$i] .= " $#$data_ref\:$linenum";
                $start =~ s/^.*://;
                if( ! $G_WARN_P_STOP )
                  {
                    $G_WARN_P_STOP = 1;
                    &print_error( "Corresponding 'stop' not found for block ".
                                  "[$block]",
                                  "Setting to last sampling",
                                  "Ignoring future warnings of this type",
                                  "File: $filename:$start", 0 );
                  }
              }
          }
      }
    #.................................
    #...check for field consistency...
    #.................................
    foreach $field ( @G_FNREQ )
      {
        if( ! defined( $data[0]{$field} ) )
          {
            $ierr = 1;
            &print_error( "Bad data file [$filename]",
                          "Could not locate field [$field]",
                          $ierr );
            exit( $ierr );
          }
      }
    close( FILE );
  }
#............................................................................
#...Name
#...====
#... get_results
#...
#...Purpose
#...=======
#... Get results of a block
#...
#...Arguments
#...=========
#... start_stop_ref    Intent: in
#...                   Perl type: reference to hash
#...                   Start/stop data indices per instance of block
#...
#... data_ref          Intent: in
#...                   Perl type: reference to hash
#...                   For each line, contains all field data
#...
#... filenum           Intent: in
#...                   Perl type: scalar
#...                   The current file number
#...
#... results_ref       Intent: out
#...                   Perl type: reference to hash
#...                   The results of each block
#...
#... results_all_ref   Intent: out
#...                   Perl type: reference to hash
#...                   The results of the results for the blocks
#...
#...Program Flow
#...============
#... 1) foreach block
#... 1.1) foreach instance of the block
#... 1.1.1) get start/stop indices in data
#... 1.1.2) get some values for this block
#... 1.1.3) foreach data line in this instance, get some other values
#............................................................................
sub get_results
  {
    my(
       $start_stop_ref,
       $data_ref,
       $filenum,
       $results_ref,
       $results_all_ref
      ) = @_;
    my(
       $block, # the current block
       $i, # loop var
       $j, # loop var
       $max_m, # current maximum memory - used for finding the actual max
       $start, # start/stop pair
       $stop, # start/stop pair
       $tmp, # tmp var
       $tmp1, # tmp var
      );
    #........................
    #...process each block...
    #........................
    undef( %$results_ref );
    foreach $block ( sort keys %$start_stop_ref )
      {
        $$results_ref{$block}{num} = 0;
        #........................................
        #...process each instance of the block...
        #........................................
        for( $j = 0; $j <= $#{$$start_stop_ref{$block}}; $j++ )
          {
            ($start, $stop) = split( / /, $$start_stop_ref{$block}[$j] );
            #.................................
            #...strip off file line numbers...
            #.................................
            $start =~ s/:.*//;
            $stop =~ s/:.*//;
            #.................................
            #...data for this results block...
            #.................................
            $$results_ref{$block}{num}++;
            push( @{$$results_ref{$block}{time}},
                  $$data_ref[$start]{$G_FN_EPOCH_TIME} );
            push( @{$$results_ref{$block}{filenum}}, $filenum );
            push( @{$$results_ref{$block}{delta_m}},
                  $$data_ref[$stop]{$G_FN_PSIZE} - $$data_ref[$start]{$G_FN_PSIZE} );
            push( @{$$results_ref{$block}{delta_t}},
                  $$data_ref[$stop]{$G_FN_EPOCH_TIME}-$$data_ref[$start]{$G_FN_EPOCH_TIME} );
            #....................................................
            #...loop through this data block and update fields...
            #....................................................
            $max_m = -1e99;
            for( $i = $start; $i <= $stop; $i++ )
              {
                $tmp = $$data_ref[$i]{$G_FN_PSIZE} - $$data_ref[$start]{$G_FN_PSIZE};
                if( $tmp > $max_m )
                  {
                    $max_m = $tmp;
                  }
              }
            push( @{$$results_ref{$block}{max_m}}, $max_m );
          }
        #..............................................
        #...DONE: process each instance of the block...
        #..............................................
      }
    #..............................
    #...DONE: process each block...
    #..............................
    #............................
    #...results -> results_all...
    #............................
    foreach $block ( keys %$results_ref )
      {
        if( ! defined( $$results_all_ref{$block} ) )
          {
            $$results_all_ref{$block}{num}     = 0;
            $$results_all_ref{$block}{delta_m} = 0;
            $$results_all_ref{$block}{delta_t} = 0;
            $$results_all_ref{$block}{max_m}   = -1e99;
            $$results_all_ref{$block}{max_t}   = -1e99;
          }
        $$results_all_ref{$block}{num} += $$results_ref{$block}{num};
        for( $i = 0; $i <= $#{$$results_ref{$block}{time}}; $i++ )
          {
            $$results_all_ref{$block}{delta_m}  +=
              $$results_ref{$block}{delta_m}[$i];
            $$results_all_ref{$block}{delta_t}  +=
              $$results_ref{$block}{delta_t}[$i];
            if( $$results_all_ref{$block}{max_m} <
                $$results_ref{$block}{max_m}[$i] )
              {
                $$results_all_ref{$block}{max_m} =
                  $$results_ref{$block}{max_m}[$i];
              }
            if( $$results_all_ref{$block}{max_t} <
                $$results_ref{$block}{delta_t}[$i] )
              {
                $$results_all_ref{$block}{max_t} =
                  $$results_ref{$block}{delta_t}[$i];
              }
          }
      }
    #..................................
    #...DONE: results -> results_all...
    #..................................
  }
#............................................................................
#...Name
#...====
#... get_results_trace
#...
#...Purpose
#...=======
#... Get results of trace data (and accumulate to trace_data_all
#...
#...Arguments
#...=========
#... trace_data_ref     Intent: inout
#...                    Perl type: reference to hash
#...                    Info about the trace data.
#...
#... data_ref           Intent: in
#...                    Perl type: reference to hash
#...                    For each line, contains all field data
#...
#... filenum            Intent: in
#...                    Perl type: scalar
#...                    The current file number
#...
#... trace_data_all_ref Intent: out
#...                    Perl type: reference to hash
#...                    The results of the results for the blocks
#...
#...Program Flow
#...============
#... 1) foreach block
#... 1.1) foreach instance of the block
#... 1.1.1) get start/stop indices in data
#... 1.1.2) get some values for this block
#... 1.1.3) foreach data line in this instance, get some other values
#............................................................................
sub get_results_trace
  {
    my(
       $trace_data_ref,
       $data_ref,
       $filenum,
       $trace_data_all_ref
      ) = @_;
    my(
       $block, # the current block
       $i, # loop var
       $j, # loop var
       $max_m, # current maximum memory - used for finding the actual max
       $start, # start/stop pair
       $stop, # start/stop pair
       $tmp, # tmp var
       $tmp1, # tmp var
      );
    #...............................
    #...get info about this block...
    #...............................
    $$trace_data_ref{num} = $#{$$trace_data_ref{start_stop}} + 1;
    #........................................
    #...process each instance of the block...
    #........................................
    for( $j = 0; $j < $$trace_data_ref{num}; $j++ )
      {
        ($start, $stop) = split( / /, $$trace_data_ref{start_stop}[$j] );
        if( ! defined( $stop ) )
          {
            $stop = "$#$data_ref\:-1";
          }
        #.................................
        #...strip off file line numbers...
        #.................................
        $start =~ s/:.*//;
        $stop =~ s/:.*//;
        #.................................
        #...data for this results block...
        #.................................
        push( @{$$trace_data_ref{time}},
              $$data_ref[$start]{$G_FN_EPOCH_TIME} );
        push( @{$$trace_data_ref{filenum}}, $filenum );
        push( @{$$trace_data_ref{delta_m}},
              $$data_ref[$stop]{$G_FN_PSIZE} - $$data_ref[$start]{$G_FN_PSIZE} );
        push( @{$$trace_data_ref{delta_t}},
              $$data_ref[$stop]{$G_FN_EPOCH_TIME}-$$data_ref[$start]{$G_FN_EPOCH_TIME} );
        #....................................................
        #...loop through this data block and update fields...
        #....................................................
        $max_m = -1e99;
        for( $i = $start; $i <= $stop; $i++ )
          {
            $tmp = $$data_ref[$i]{$G_FN_PSIZE} - $$data_ref[$start]{$G_FN_PSIZE};
            if( $tmp > $max_m )
              {
                $max_m = $tmp;
              }
          }
        push( @{$$trace_data_ref{max_m}}, $max_m );
      }
    #.....................................
    #...DONE: get info about this block...
    #.....................................
    #......................................
    #...add block info to trace_data_all...
    #......................................
    if( ! defined( $$trace_data_all_ref{num} ) )
      {
        $$trace_data_all_ref{num}     = 0;
        $$trace_data_all_ref{delta_m} = 0;
        $$trace_data_all_ref{delta_t} = 0;
        $$trace_data_all_ref{max_m}   = -1e99;
        $$trace_data_all_ref{max_t}   = -1e99;
      }
    $$trace_data_all_ref{num} += $$trace_data_ref{num};
    for( $j = 0; $j < $$trace_data_ref{num}; $j++ )
      {
        $$trace_data_all_ref{delta_m} += $$trace_data_ref{delta_m}[$j];
        $$trace_data_all_ref{delta_t} += $$trace_data_ref{delta_t}[$j];
        if( $$trace_data_all_ref{max_m} < $$trace_data_ref{delta_m}[$j] )
          {
            $$trace_data_all_ref{max_m} = $$trace_data_ref{delta_m}[$j];
          }
        if( $$trace_data_all_ref{max_t} < $$trace_data_ref{delta_t}[$j] )
          {
            $$trace_data_all_ref{max_t} = $$trace_data_ref{delta_t}[$j];
          }
      }
    #............................................
    #...DONE: add block info to trace_data_all...
    #............................................
    #......................................
    #...recursive call to other children...
    #......................................
    if( defined( $$trace_data_ref{child} ) )
      {
        foreach $block( keys %{$$trace_data_ref{child}} )
          {
            &get_results_trace( \%{$$trace_data_ref{child}{$block}},
                                $data_ref, $filenum,
                                \%{$$trace_data_all_ref{child}{$block}} );
          }
      }
    #........................
    #...get exclusive data...
    #........................
    $$trace_data_all_ref{delta_m_exc} = $$trace_data_all_ref{delta_m};
    $$trace_data_all_ref{delta_t_exc} = $$trace_data_all_ref{delta_t};
    if( defined( $$trace_data_ref{child} ) )
      {
        foreach $block( keys %{$$trace_data_ref{child}} )
          {
            $$trace_data_all_ref{delta_m_exc} -=
              $$trace_data_all_ref{child}{$block}{delta_m};
            $$trace_data_all_ref{delta_t_exc} -=
              $$trace_data_all_ref{child}{$block}{delta_t};
          }
      }
  }
#............................................................................
#...Name
#...====
#... trace_data_all_finish
#...
#...Purpose
#...=======
#... Finish off some stats when gotten all trace data.
#...
#...Arguments
#...=========
#... trace_data_all_root_ref   Intent: in
#...                           Perl type: reference to hash
#...                           The top level trace_data hash
#...
#... trace_data_all_ref        Intent: inout
#...                           Perl type: reference to hash
#...                           trace_data hash
#...
#...Program Flow
#...============
#... 1) finish off data
#... 2) call recursively for children
#............................................................................
sub trace_data_all_finish
  {
    my(
       $trace_data_all_root_ref,
       $trace_data_all_ref,
      ) = @_;
    my(
       $block, # the block
       $den_m, # m denominator
       $den_t, # t denominator
      );
    $den_m = $$trace_data_all_root_ref{delta_m};
    if( $den_m == 0 )
      {
        $$trace_data_all_ref{delta_m_pct} = 0;
      }
    else
      {
        $$trace_data_all_ref{delta_m_pct} =
          ( $$trace_data_all_ref{delta_m} / $den_m ) * 100.0;
      }
    $den_t = $$trace_data_all_root_ref{delta_t};
    if( $den_t == 0 )
      {
        $$trace_data_all_ref{delta_t_pct} = 0;
      }
    else
      {
        $$trace_data_all_ref{delta_t_pct} =
          ( $$trace_data_all_ref{delta_t} / $den_t ) * 100.0;
      }
    $den_m = $$trace_data_all_root_ref{delta_m};
    if( $den_m == 0 )
      {
        $$trace_data_all_ref{delta_m_exc_pct} = 0;
      }
    else
      {
        $$trace_data_all_ref{delta_m_exc_pct} =
          ( $$trace_data_all_ref{delta_m_exc} / $den_m ) * 100.0;
      }
    $den_t = $$trace_data_all_root_ref{delta_t};
    if( $den_t == 0 )
      {
        $$trace_data_all_ref{delta_t_exc_pct} = 0;
      }
    else
      {
        $$trace_data_all_ref{delta_t_exc_pct} =
          ( $$trace_data_all_ref{delta_t_exc} / $den_t ) * 100.0;
      }
    if( defined( $$trace_data_all_ref{child} ) )
      {
        foreach $block( keys %{$$trace_data_all_ref{child}} )
          {
            &trace_data_all_finish( $trace_data_all_root_ref,
                                    \%{$$trace_data_all_ref{child}{$block}} );
          }
      }
  }
#............................................................................
#...Name
#...====
#... print_results_all
#...
#...Purpose
#...=======
#... Print the final results
#...
#...Arguments
#...=========
#... results_all_ref   Intent: in
#...                   Perl type: reference to hash
#...                   The results of the results for the blocks
#...
#... cmd_ref           Intent: in
#...                   Perl type: reference to hash
#...                   Reference to command line arguments hash.
#...
#...Program Flow
#...============
#... 1) print header
#... 2) print results from each block (with final final results last)
#............................................................................
sub print_results_all
  {
    my(
       $results_all_ref,
       $cmd_ref
      ) = @_;
    my(
       $block, # the block
       $filtered, # if filtering out
       $m_avg, # average m
       $t_avg, # average t
      );
    printf "\n";
    printf "Inclusive Total:\n";
    printf( "%10s %10s %10s %10s %10s %10s %10s %s\n",
            "Delta Mem",
            "average",
            "max",
            "Delta Time",
            "average",
            "max",
            "# Blocks",
            "Block"
          );
    printf( "%10s %10s %10s %10s %10s %10s %10s %s\n",
            "---------",
            "-------",
            "---",
            "----------",
            "-------",
            "---",
            "--------",
            "-----"
          );
    foreach $block ( grep( !/^$G_B_FINAL_RES$/, sort( keys %$results_all_ref ) ),
                     "$G_B_FINAL_RES" )
      {
        #......................
        #...next if filtered...
        #......................
        $filtered = 1;
        if( $block eq "$G_B_FINAL_RES" ||
            ! defined( $$cmd_ref{$G_CMD_FILTER}{memory} ) ||
            $$results_all_ref{$block}{delta_m} != 0 )
          {
            $filtered = 0;
          }
        if( $filtered )
          {
            next;
          }
        #..................
        #...not filtered...
        #..................
        if( $block eq "$G_B_FINAL_RES" )
          {
            print "\n";
          }
        $avg_m = $results_all{$block}{delta_m}/$results_all{$block}{num};
        $avg_t = $results_all{$block}{delta_t}/$results_all{$block}{num};
        printf( "%10.2e %10.2e %10.2e %10.2e %10.2e %10.2e %10.2e %s\n",
                $$results_all_ref{$block}{delta_m},
                $avg_m,
                $$results_all_ref{$block}{max_m},
                $$results_all_ref{$block}{delta_t},
                $avg_t,
                $$results_all_ref{$block}{max_t},
                $$results_all_ref{$block}{num},
                $block
              );
      }
    print "\n";
  }
#............................................................................
#...Name
#...====
#... print_trace_data_all
#...
#...Purpose
#...=======
#... Print the final results
#...
#...Arguments
#...=========
#... trace_data_all_ref   Intent: in
#...                      Perl type: reference to hash
#...                      The results all the trace_data from all files
#...
#... cmd_ref              Intent: in
#...                      Perl type: reference to hash
#...                      Reference to command line arguments hash.
#...
#...Program Flow
#...============
#... 1) print header
#... 2) print results from each block (with final final results last)
#............................................................................
sub print_trace_data_all
  {
    my(
       $trace_data_all_ref,
       $cmd_ref
      ) = @_;
    printf "\n";
    printf "Inclusive Tree:\n";
    printf( "%10s %10s %10s %6s %10s %10s %10s %6s %10s %s\n",
            "Delta Mem",
            "average",
            "max",
            "%",
            "Delta Time",
            "average",
            "max",
            "%",
            "# Blocks",
            "Block"
          );
    printf( "%10s %10s %10s %6s %10s %10s %10s %6s %10s %s\n",
            "---------",
            "-------",
            "---",
            "-",
            "----------",
            "-------",
            "---",
            "-",
            "--------",
            "-----"
          );
    &print_trace_data_all_block( $trace_data_all_ref, $cmd_ref,
                                 $G_B_FINAL_RES, 0 );
    printf "\n";
    printf "Exclusive Tree:\n";
    printf( "%10s %10s %6s %10s %10s %6s %10s %s\n",
            "Delta Mem",
            "average",
            "%",
            "Delta Time",
            "average",
            "%",
            "# Blocks",
            "Block"
          );
    printf( "%10s %10s %6s %10s %10s %6s %10s %s\n",
            "---------",
            "-------",
            "-",
            "----------",
            "-------",
            "-",
            "--------",
            "-----"
          );
    &print_trace_data_all_block_exc( $trace_data_all_ref, $cmd_ref,
                                     $G_B_FINAL_RES, 0 );
  }
#............................................................................
#...Name
#...====
#... print_trace_data_all_block
#...
#...Purpose
#...=======
#... Print the final results inclusive
#...
#...Arguments
#...=========
#... trace_data_all_ref   Intent: in
#...                      Perl type: reference to hash
#...                      The results all the trace_data from all files
#...
#... cmd_ref              Intent: in
#...                      Perl type: reference to hash
#...                      Reference to command line arguments hash.
#...
#... name                 Intent: in
#...                      Perl type: scalar
#...                      The block name (for printing)
#...
#... level                Intent: in
#...                      Perl type: scalar
#...                      Level in the tree (used for indenting name)
#...
#...Program Flow
#...============
#... 1) print results from this block
#... 2) call recursively for children
#............................................................................
sub print_trace_data_all_block
  {
    my(
       $trace_data_all_ref,
       $cmd_ref,
       $name,
       $level,
      ) = @_;
    my(
       $block, # the block
       $filtered, # if filtering out
       $i, # loop var
       $m_avg, # average memory
       $spaces, # spaces to print before name
       $t_avg, # average time
      );
    #......................
    #...next if filtered...
    #......................
    $filtered = 1;
    if( ! defined( $$cmd_ref{$G_CMD_FILTER}{memory} ) ||
        $$trace_data_all_ref{delta_m} != 0 )
      {
        $filtered = 0;
      }
    if( $filtered == 0 )
      {
        $spaces = "";
        for( $i = 0; $i < $level; $i++ )
          {
            $spaces .= " ";
          }
        $m_avg = $$trace_data_all_ref{delta_m} / $$trace_data_all_ref{num};
        $t_avg = $$trace_data_all_ref{delta_t} / $$trace_data_all_ref{num};
        printf( "%10.2e %10.2e %10.2e %6.1f %10.2e %10.2e %10.2e %6.1f %10.2e %s%s\n",
                $$trace_data_all_ref{delta_m},
                $m_avg,
                $$trace_data_all_ref{max_m},
                $$trace_data_all_ref{delta_m_pct},
                $$trace_data_all_ref{delta_t},
                $t_avg,
                $$trace_data_all_ref{max_t},
                $$trace_data_all_ref{delta_t_pct},
                $$trace_data_all_ref{num},
                $spaces,
                $name
              );
        if( defined( $$trace_data_all_ref{child} ) )
          {
            foreach $block( sort keys %{$$trace_data_all_ref{child}} )
              {
                &print_trace_data_all_block( \%{$$trace_data_all_ref{child}{$block}},
                                             $cmd_ref, $block, $level + 1 );
              }
          }
      }
  }
#............................................................................
#...Name
#...====
#... print_trace_data_all_block_exc
#...
#...Purpose
#...=======
#... Print the final results exclusive
#...
#...Arguments
#...=========
#... trace_data_all_ref   Intent: in
#...                      Perl type: reference to hash
#...                      The results all the trace_data from all files
#...
#... cmd_ref              Intent: in
#...                      Perl type: reference to hash
#...                      Reference to command line arguments hash.
#...
#... name                 Intent: in
#...                      Perl type: scalar
#...                      The block name (for printing)
#...
#... level                Intent: in
#...                      Perl type: scalar
#...                      Level in the tree (used for indenting name)
#...
#...Program Flow
#...============
#... 1) print results from this block
#... 2) call recursively for children
#............................................................................
sub print_trace_data_all_block_exc
  {
    my(
       $trace_data_all_ref,
       $cmd_ref,
       $name,
       $level,
      ) = @_;
    my(
       $block, # the block
       $filtered, # if filtering out
       $i, # loop var
       $m_avg, # average memory
       $spaces, # spaces to print before name
       $t_avg, # average time
      );
    #......................
    #...next if filtered...
    #......................
    $filtered = 1;
    if( ! defined( $$cmd_ref{$G_CMD_FILTER}{memory} ) ||
        $$trace_data_all_ref{delta_m} != 0 )
      {
        $filtered = 0;
      }
    if( $filtered == 0 )
      {
        $spaces = "";
        for( $i = 0; $i < $level; $i++ )
          {
            $spaces .= " ";
          }
        $m_avg = $$trace_data_all_ref{delta_m_exc} / $$trace_data_all_ref{num};
        $t_avg = $$trace_data_all_ref{delta_t_exc} / $$trace_data_all_ref{num};
        printf( "%10.2e %10.2e %6.1f %10.2e %10.2e %6.1f %10.2e %s%s\n",
                $$trace_data_all_ref{delta_m_exc},
                $m_avg,
                $$trace_data_all_ref{delta_m_exc_pct},
                $$trace_data_all_ref{delta_t_exc},
                $t_avg,
                $$trace_data_all_ref{delta_t_exc_pct},
                $$trace_data_all_ref{num},
                $spaces,
                $name
              );
        if( defined( $$trace_data_all_ref{child} ) )
          {
            foreach $block( sort keys %{$$trace_data_all_ref{child}} )
              {
                &print_trace_data_all_block_exc( \%{$$trace_data_all_ref{child}{$block}},
                                                 $cmd_ref, $block,
                                                 $level + 1 );
              }
          }
      }
  }
#............................................................................
#...Name
#...====
#... print_data_info
#...
#...Purpose
#...=======
#... Create the file(s) gnuplot will use and stuff gnuplot_info with
#... info needed to do plots.
#...
#...Arguments
#...=========
#... data_ref          Intent: in
#...                   Perl type: reference to hash
#...                   For each line, contains all field data
#...
#... field_info        Intent: in
#...                   Perl type: reference to hash
#...                   Contains info about the fields
#...
#... filenum           Intent: in
#...                   Perl type: scalar
#...                   The current file number
#...
#... filename          Intent: in
#...                   Perl type: scalar
#...                   The name of the file
#...
#... gnuplot_info      Intent: inout
#...                   Perl type: reference to hash
#...                   Contains info needed for making gnuplot's of the data
#...
#... fh                Intent: inout
#...                   Perl type: glob of filehandle
#...                   The filhandle to print to
#...
#... ierr              Intent: out
#...                   Perl type: scalar
#...                   error return value
#...
#...Program Flow
#...============
#... 1) original data header + gnuplot_info
#... 2) original data
#... 3) process each block data
#... 3.1) block data header + gnuplot_info
#... 3.2) block data
#............................................................................
sub print_data_info
  {
    my(
       $data_ref,
       $field_info_ref,
       $filenum,
       $filename,
       $gnuplot_info_ref,
       $fh
      ) = @_;
    my(
       $block, # current block
       $using, # keeping track of gnuplot value
       $field, # current field
       $filename_dat, # name of the output data file
       $i, # loop var
       $ierr, # error return value
       $index, # gnuplot index
      );
    $ierr = 0;
    $index = -1;
    #...............
    #...open file...
    #...............
    $filename_dat = "$filename.dat";
    if( ! open( FILE, ">$filename_dat" ) )
      {
        $ierr = 1;
        &print_error( "Cannot open output data file [$filename_dat]",
                      -1 );
        return( $ierr );
      }
    #.........................................
    #...original data header + gnuplot_info...
    #.........................................
    $index++;
    print FILE "\n\n# $filename\[$filenum\]: $G_B_ORIG_DATA\n";
    printf FILE "# %17s", "Time";
    $using = 1;
    foreach $field ( @{$$field_info_ref{"$G_B_ORIG_DATA"}{order}} )
      {
        #...skip if field does not exist...
        if( !defined( $$data_ref[0]{$field} ) )
          {
            next;
          }
        #...skip all fields from here on...
        if( $field eq "$G_FN_MMEMSIZE" )
          {
            last;
          }
        #...time already taken care of...
        if( $field eq "$G_FN_EPOCH_TIME" ||
            $field eq "$G_FN_PWALLTIME" )
          {
            next;
          }
        #...print the field...
        printf FILE "%10s", $field;
        #............................
        #...save gnuplot_info data...
        #............................
        $using++;
        push( @{$$gnuplot_info_ref{"$G_B_ORIG_DATA"}{$field}{index}}, $index );
        push( @{$$gnuplot_info_ref{"$G_B_ORIG_DATA"}{$field}{using}}, $using );
        push( @{$$gnuplot_info_ref{"$G_B_ORIG_DATA"}{$field}{filenum}}, $filenum );
        push( @{$$gnuplot_info_ref{"$G_B_ORIG_DATA"}{$field}{filename}}, $filename_dat );
      }
    print FILE "\n";
    #...................
    #...original data...
    #...................
    for( $i = 0; $i <= $#$data_ref; $i++ )
      {
        printf FILE "  %17.6f", $$data_ref[$i]{$G_FN_EPOCH_TIME};
        foreach $field ( @{$$field_info_ref{"$G_B_ORIG_DATA"}{order}} )
          {
            #...skip if field does not exist...
            if( !defined( $$data_ref[0]{$field} ) )
              {
                next;
              }
            #...skip all fields from here on...
            if( $field eq "$G_FN_MMEMSIZE" )
              {
                last;
              }
            #...time already taken care of...
            if( $field eq "$G_FN_EPOCH_TIME" ||
                $field eq "$G_FN_PWALLTIME" )
              {
                next;
              }
            #...print the field...
            printf FILE "%10.2e", $$data_ref[$i]{$field};
          }
        print FILE "\n";
      }
    #.............................
    #...process each block data...
    #.............................
    foreach $block ( grep( !/^$G_B_FINAL_RES$/, sort keys %results ),
                     "$G_B_FINAL_RES" )
      {
        #......................................
        #...block data header + gnuplot_info...
        #......................................
        $index++;
        printf FILE "\n\n# $filename\[$filenum\]: $block\n";
        printf FILE "# %17s %7s", "time", "filenum";
        $using = 2;
        foreach $field ( @{$$field_info_ref{block}{order}} )
          {
            #...skip if field does not exist...
            if( !defined( $results{$block}{$field} ) )
              {
                next;
              }
            #...time already taken care of, and num not interested in...
            if( $field eq "time" || $field eq "num" )
              {
                next;
              }
            printf FILE "%10s", $field;
            #............................
            #...save gnuplot_info data...
            #............................
            $using++;
            push( @{$$gnuplot_info_ref{$block}{$field}{index}}, $index );
            push( @{$$gnuplot_info_ref{$block}{$field}{using}}, $using );
            push( @{$$gnuplot_info_ref{$block}{$field}{filenum}}, $filenum );
            push( @{$$gnuplot_info_ref{$block}{$field}{filename}}, $filename_dat );
          }
        print FILE "\n";
        #................
        #...block data...
        #................
        for( $i = 0; $i <= $#{$results{$block}{time}}; $i++ )
          {
            printf FILE "  %17.6f", $results{$block}{time}[$i];
            printf FILE " %7d", $results{$block}{filenum}[$i];
            foreach $field ( @{$$field_info_ref{block}{order}} )
              {
                #...skip if field does not exist...
                if( !defined( $results{$block}{$field} ) )
                  {
                    next;
                  }
                #...time already taken care of, and num not interested in...
                if( $field eq "time" || $field eq "num" )
                  {
                    next;
                  }
                printf FILE "%10.2e", $results{$block}{$field}[$i];
              }
            print FILE "\n";
          }
      }
    #...................................
    #...DONE: process each block data...
    #...................................
    #................
    #...close file...
    #................
    close( FILE );
    return( $ierr );
  }
#............................................................................
#...Name
#...====
#... plot
#...
#...Purpose
#...=======
#... Create plots of the data
#...
#...Arguments
#...=========
#... gnuplot_info_ref  Intent: in
#...                   Perl type: reference to hash
#...                   reference to %gnuplot_info - contains
#...                   data like index, using, filenum, ...
#...
#... field_info_ref    Intent: in
#...                   Perl type: reference to hash
#...                   reference to %field_info - contains
#...                   data like order of fields and their descriptions.
#...
#... filenames_ref     Intent: in
#...                   Perl type: reference to array
#...                   ordered list of files (for making key)
#...
#... maxtime           Intent: in
#...                   Perl type: scalar
#...                   The maximum time (used for setting yrange)
#...
#... results_all_ref   Intent: in
#...                   Perl type: reference to hash
#...                   The results of the results for the blocks
#...
#... cmd_ref           Intent: in
#...                   Perl type: reference to hash
#...                   Reference to command line arguments hash.
#...
#...Program Flow
#...============
#... 1) find execs
#... 2) init gnuplot_cmd
#... 3) process each block (new multiplot)
#... 3.1) process each field (plot in a multiplot)
#... 4) print key
#... 5) run gnuplot
#............................................................................
sub plot
  {
    my(
       $gnuplot_info_ref,
       $field_info_ref,
       $filenames_ref,
       $maxtime,
       $results_all_ref,
       $cmd_ref
      ) = @_;
    my(
       $arch, # architecture
       $block, # the block
       @fields, # all the fields that could be used
       $field_type, # key into $field_info_ref
       $file_range, # the xrange when plotting x with file numbers
       $filtered, # if filtering out
       $gnuplot, # where gnuplot is
       $gnuplot_cmd, # the gnuplot command to print to file
       $gnuplot_size, # size for this plot
       $gnuplot_origin, # current starting point for this plot
       $i, # loop var
       $lt, # line type
       $num_plots_page, # number of plots per page
       $output_file_ps, # output file
       $output_file_pdf, # output file
       $output_file_cmd, # output file
       $plot_num, # current plot number for this page
       $ps2pdf, # where ps2pdf is
       $scale_factor, # for scaling of plots given 1st plot/page has labels
       $title, # title of a plot
      );
    #..........
    #...init...
    #..........
    $file_range = $#$filenames_ref + 2;
    $num_plots_page = 4;
    $scale_factor = 1 + .75*($num_plots_page/4); # scaled for font size 10
    $output_file_ps  = "procmon.ps";
    $output_file_pdf = "procmon.pdf";
    $output_file_cmd = "procmon.cmd";
    #................
    #...find execs...
    #................
    $arch = `uname -a 2>&1`;
    if( $arch =~ /alpha/i )
      {
        $ENV{PATH} = "/usr/local/gnu/bin:$ENV{PATH}";
        $ENV{PATH} = "/usr/projects/ups/ghostscript/alpha/bin:$ENV{PATH}";
      }
    if( $arch =~ /irix/i )
      {
        $ENV{PATH} = "/usr/local/gnu/bin:$ENV{PATH}";
        $ENV{PATH} = "/usr/projects/ups/ghostscript/irix/bin:$ENV{PATH}";
      }
    if( $arch =~ /linux/i )
      {
        $ENV{PATH} = "/usr/projects/ups/ghostscript/linux/bin:$ENV{PATH}";
      }
    $gnuplot = &which_exec( "gnuplot" );
    if( $gnuplot !~ /\S/ )
      {
        return;
      }
    $ps2pdf = &which_exec( "ps2pdf" );
    #........................
    #...create gnuplot_cmd...
    #........................
    $gnuplot_cmd = "";
    $gnuplot_cmd .= "# \n";
    $gnuplot_cmd .= "# This file is to be processed by gnuplot.\n";
    $gnuplot_cmd .= "# The data file(s) from PROCMON are read and\n";
    $gnuplot_cmd .= "# $output_file_ps is created.\n";
    $gnuplot_cmd .= "# \n";
    $gnuplot_cmd .= "set output '$output_file_ps'\n";
    $gnuplot_cmd .= "set terminal postscript portrait color 'Times-Roman' 10\n";
    #........................
    #...process each block...
    #........................
    foreach $block ( "$G_B_ORIG_DATA",
                     grep( !/^($G_B_ORIG_DATA|$G_B_FINAL_RES)$/,
                           sort keys %$gnuplot_info_ref ),
                     "$G_B_FINAL_RES" )
      {
        #......................
        #...next if filtered...
        #......................
        $filtered = 1;
        if( $block eq "$G_B_FINAL_RES" ||
            $block eq "$G_B_ORIG_DATA" ||
            ! defined( $$cmd_ref{$G_CMD_FILTER}{memory} ) ||
            $$results_all_ref{$block}{delta_m} != 0 )
          {
            $filtered = 0;
          }
        if( $filtered )
          {
            next;
          }
        #..................
        #...not filtered...
        #..................
        if( $block eq "$G_B_ORIG_DATA" )
          {
            $field_type = $block;
          }
        else
          {
            $field_type = "block";
          }
        @fields = @{$$field_info_ref{$field_type}{order}};
        #....................................................
        #...init for this block (and clear any multiplots)...
        #....................................................
        $gnuplot_cmd .= "set nomultiplot\n";
        $plot_num = 0;
        #........................
        #...process each field...
        #........................
        foreach $field ( @fields )
          {
            if( ! defined( $$gnuplot_info_ref{$block}{$field} ) )
              {
                next;
              }
            $title = "$block: $field = ".
              "$$field_info_ref{$field_type}{def}{$field}";
            $gnuplot_cmd .= "#\n";
            $gnuplot_cmd .= "# Plotting [$title] ";
            if( $plot_num == 0 )
              {
                $gnuplot_cmd .= "on a new page.\n";
                $gnuplot_cmd .= "# \n";
                $gnuplot_cmd .= "set multiplot\n";
                $gnuplot_cmd .= "set nokey\n";
                $gnuplot_cmd .= "set grid x2tics ytics\n";
                $gnuplot_cmd .= "set format x ''\n";
                $gnuplot_cmd .= "set x2tics\n";
                if( $block ne "$G_B_FINAL_RES" )
                  {
                    $gnuplot_cmd .= "set x2label 'Time (secs)'\n";
                    $gnuplot_cmd .= 'set format x2 "%17f"'."\n";
                    $gnuplot_cmd .= "set  xrange [0:$maxtime]\n";
                    $gnuplot_cmd .= "set x2range [0:$maxtime]\n";
                  }
                else
                  {
                    $gnuplot_cmd .= "set x2label 'File Number'\n";
                    $gnuplot_cmd .= 'set format x2 "%.0f"'."\n";
                    $gnuplot_cmd .= "set  xrange [0:$file_range]\n";
                    $gnuplot_cmd .= "set x2range [0:$file_range]\n";
                  }
                $gnuplot_cmd .= "set x2tics rotate\n";
                $gnuplot_cmd .= "set ytics\n";
                $gnuplot_cmd .= "set ylabel 'Value'\n";
                $gnuplot_cmd .= 'set format y "%12.5e"'."\n";
                $gnuplot_size = $scale_factor/($scale_factor+$num_plots_page-1);
                $gnuplot_origin = 1 - $gnuplot_size;
              }
            else
              {
                $gnuplot_cmd .= "on the same page.\n";
                $gnuplot_cmd .= "# \n";
                $gnuplot_cmd .= "set x2label ''\n";
                $gnuplot_cmd .= 'set format x2 ""'."\n";
                $gnuplot_size = 1/($scale_factor+$num_plots_page-1);
                $gnuplot_origin -= $gnuplot_size;
              }
            $gnuplot_cmd .= sprintf( "set size 1, %.4f\n",
                                     abs($gnuplot_size) );
            $gnuplot_cmd .= sprintf( "set origin 0, %.4f\n",
                                     abs($gnuplot_origin) );
            $gnuplot_cmd .=
              "set title '$title'\n";
            $gnuplot_cmd .= "plot ";
            for( $i = 0; $i <= $#{$$gnuplot_info_ref{$block}{$field}{index}};
                 $i++ )
              {
                $gnuplot_cmd .= "'$$gnuplot_info_ref{$block}{$field}{filename}[$i]' ";
                $gnuplot_cmd .= "index ".
                  $$gnuplot_info_ref{$block}{$field}{index}[$i]." ";
                if( $block ne "$G_B_FINAL_RES" )
                  {
                    $gnuplot_cmd .= "using 1:".
                      $$gnuplot_info_ref{$block}{$field}{using}[$i]." ";
                  }
                else
                  {
                    $gnuplot_cmd .= "using 2:".
                      $$gnuplot_info_ref{$block}{$field}{using}[$i]." ";
                  }
                $gnuplot_cmd .= "notitle ";
                $lt = ($$gnuplot_info_ref{$block}{$field}{filenum}[$i]-1)%64+1;
                if( $block eq "$G_B_ORIG_DATA" )
                  {
                    $gnuplot_cmd .= "with lines lt $lt lw 3";
                  }
                elsif( $block eq "$G_B_FINAL_RES" )
                  {
                    $gnuplot_cmd .= "with impulses lt $lt lw 4 ";
                  }
                else
                  {
                    $gnuplot_cmd .= "with linespoints lt $lt ";
                  }
                $gnuplot_cmd .= ", ";
              }
            $gnuplot_cmd =~ s/, $/\n/;
            $plot_num++;
            if( $plot_num >= $num_plots_page )
              {
                $plot_num = 0;
                $gnuplot_cmd .= "set nomultiplot\n";
              }
          }
        #..............................
        #...DONE: process each field...
        #..............................
        #....................................
        #...finish any remaining multiplot...
        #....................................
        $gnuplot_cmd .= "set nomultiplot\n";
      }
    #..............................
    #...DONE: process each block...
    #..............................
    #...........................
    #...make final plot a key...
    #...........................
    $gnuplot_cmd .= "# \n";
    $gnuplot_cmd .= "# Print Datafile Key\n";
    $gnuplot_cmd .= "# \n";
    $gnuplot_cmd .= "set title ''\n";
    $gnuplot_cmd .= "set nogrid\n";
    $gnuplot_cmd .= "set noborder\n";
    $gnuplot_cmd .= "set noxtics\n";
    $gnuplot_cmd .= "set noytics\n";
    $gnuplot_cmd .= "set nox2tics\n";
    $gnuplot_cmd .= "set ylabel\n";
    $gnuplot_cmd .= "set size 1,1\n";
    $gnuplot_cmd .= "set origin 0,0\n";
    $gnuplot_cmd .= "set key left Left reverse title 'Datafile Legend' box\n";
    $gnuplot_cmd .= "plot ";
    for( $i = 0; $i <= $#$filenames_ref; $i++ )
      {
        $lt = $i%64 + 1;
	($title = $$filenames_ref[$i]) =~ s/^.*\///;
        $gnuplot_cmd .= "'-' title '$title' with lines lt $lt, ";
      }
    $gnuplot_cmd =~ s/, $/\n/;
    $gnuplot_cmd .= "0\n";
    foreach $filename ( @$filenames_ref )
      {
        $gnuplot_cmd .= "e\n";
      }
    #.................
    #...run gnuplot...
    #.................
    if( $gnuplot ne "" )
      {
        if( ! open( FILE, ">$output_file_cmd" ) )
          {
            $ierr = 1;
            &print_error(
                         "Could not open gnuplot command file [$output_file_cmd].",
                         $ierr
                        );
            exit( $ierr );
          }
        printf " Creating [$output_file_cmd]\n";
        print FILE $gnuplot_cmd;
        close FILE;
        printf " Creating [$output_file_ps]\n";
        $output = `$gnuplot $output_file_cmd > /dev/null 2>&1`;
        print $output;
        if( $ps2pdf ne "" && -T "$output_file_ps" )
          {
            printf " Creating [$output_file_pdf]\n";
            $output = `$ps2pdf $output_file_ps $output_file_pdf 2>&1`;
            print $output;
          }
      }
  }
#............................................................................
#...Name
#...====
#... print_error
#...
#...Purpose
#...=======
#... Print a standard error message.
#...
#...Arguments
#...=========
#... error_lines  Intent: in
#...              Perl type: array
#...              0: cause of error (file_name:line_number)
#...              1: explanation/fix (if there is one)
#...              2: error value
#...
#...Program Flow
#...============
#... 1) see if warning or error (last argument is 0 or not)
#... 2) Line up error lines by column
#... 3) find out who was calling this routine and the line number
#... 4) print out info
#............................................................................
sub print_error
  {
    my(
       @error_lines # what emitted the error
      ) = @_;
    my(
       $error_level,  # what is printed (eg WARNING or ERROR)
       $error_line,   # each line of input argument
       @routine_info, # var from caller
       $routine_name, # routine name
       $spaces        # spaces for lining up columns
      );
    #......................................................
    #...assign WARNING or ERROR depending on error value...
    #......................................................
    if ( "$error_lines[$#error_lines]" eq "0" )
      {
        $error_level = "**WARNING**";
        $spaces      = "           ";
      }
    else
      {
        $error_level = "**ERROR**";
        $spaces      = "         ";
      }
    #............................................................
    #...DONE: assign WARNING or ERROR depending on error value...
    #............................................................
    #.......................................
    #...init error and add argument lines...
    #.......................................
    $error_line = shift( @error_lines );
    $error_message = "\n$error_level $error_line\n";
    foreach $error_line ( @error_lines )
      {
        $error_message .= "$spaces $error_line\n";
      }
    #............................................................
    #...get routine name (special if called from main program)...
    #............................................................
    @routine_info = caller(1);
    if ( $#routine_info < 3 )
      {
        $routine_name = $0;
      }
    else
      {
        $routine_name = (caller(1))[3];
      }
    $routine_name = $routine_name.":".(caller(0))[2];
    #...........................
    #...postpend routine name...
    #...........................
    $error_message .= "$spaces $routine_name\n";
    #.................
    #...print error...
    #.................
    print STDERR $error_message;
  }
#............................................................................
#...Name
#...====
#... which_exec
#...
#...Purpose
#...=======
#... Returns value for full path to executable ("" if not found).
#...
#...Arguments
#...=========
#... exec Intent: in
#...      Perl type: string scalar
#...
#...Program Flow
#...============
#... 1) go through path to find exec
#............................................................................
sub which_exec
  {
    my( 
       $exec, # exec to find
      ) = @_;
    my(
       $ierr, # error return value
       $exec_try, # see if exec is here
       $found, # if found
       $path, # current dir in search for execs
       @paths, # list of search paths for execs
       $this_dir, # current directory
      );
    #.................................................
    #...build paths from PATH and current directory...
    #.................................................
    $path = $ENV{PATH};
    @paths = split( /:/, $ENV{PATH} );
    ($this_dir = $0) =~ s/\/[^\/]+$//;
    unshift( @paths, $this_dir );
    #.....................................
    #...loop through paths to find exec...
    #.....................................
    $found = "false";
    foreach $path (@paths)
      {
        $exec_try = "$path/$exec";
        if( -x "$exec_try" )
          {
            $found = "true";
            last;
          }
      }
    #........................................
    #...error if still could not find exec...
    #........................................
    if( $found eq "false" )
      {
        $ierr = 1;
        &print_error(
                     "Necessary executable [$exec] not found in PATH",
                     $ierr
                    );
        $exec_try = "";
      }
    return( $exec_try );
  }
#............................................................................
#...Name
#...====
#... print_perl_obj
#...
#...Purpose
#...=======
#... Print the structure of a perl object...for easier debugging.
#... Mixtures of arrays, hashes, and scalars can be printed.
#...
#... Up to a certain number of hash/array values are printed.
#... Beyond this, values are skipped (unless they are a reference
#... themselves) until the last value of the array/hash.
#...
#...Arguments
#...=========
#... $obj_ref          Intent: in
#...                   Perl type: reference to array
#...                   Reference to object to print (\%, \@, \$)
#...
#... $in_pref          Intent: in
#...                   Perl type: scalar
#...                   The preface string to name this object.
#...                   If passing \%foo, a good pref might be "foo".
#...                   if not defined, a default will be used.
#...
#... $in_max_items     Intent: in
#...                   Perl type: scalar
#...                   Maximum number of items in array/hash to print.
#...                   If negative, print all items.
#...                   If not defined, all items will be printed.
#...
#... $in_file          Intent: in
#...                   Perl type: scalar
#...                   Output file (usually just STDOUT).
#...                   on the same plot.
#...                   If not defined, STDOUT will be used.
#...
#...Program Flow
#...============
#... 1) If hash or array, recursively call this routine.
#... 2) If scalar, print the value.
#............................................................................
sub print_perl_obj
{
  my(
     $obj_ref,
     $in_pref,
     $in_max_items,
     $in_file
    ) = @_;
  my(
     $file, # file to use
     $i, # loop var
     @indices, # indices of object
     $j, # loop var
     $key, # key
     $max_items, # max items to use
     $new_pref, # new preface
     $pref, # preface to use
     $skip, # if skipping
     $skip_count, # how many skipped
     $val, # scalar value
    );
  #..........................
  #...fix non-defined vals...
  #..........................
  if( ! defined( $in_pref ) )
    {
      $pref = "var";
    }
  else
    {
      $pref = $in_pref;
    }
  if( ! defined( $in_max_items ) )
    {
      $max_items = -1;
    }
  else
    {
      $max_items = $in_max_items;
    }
  if( ! defined( $in_file ) )
    {
      $file = STDOUT;
    }
  else
    {
      $file = $in_file;
    }
  #...............................
  #...print depending upon type...
  #...............................
  if( ref( $obj_ref ) eq "ARRAY" )
  {
    if( $#{$obj_ref} < 0 )
      {
        print {$file} "$pref\[\] undefined\n";
      }
    else
      {
        $skip = 0;
        $skip_count = 0;
        for( $i = 0; $i <= $#{$obj_ref}; $i++ )
          {
            #...determine if should skip printing this one...
            if( $i > $max_items-1 && $max_items >= 0 )
              {
                $skip = 1;
              }
            $new_pref = sprintf( "%s[%s]", $pref, $i );
            $val = $$obj_ref[$i];
            if( ref( $val ) )
              {
                #...print number skipped...
                if( $skip_count > 0 )
                  {
                    print {$file} "$pref\[...$skip_count\]\n";
                    $skip_count = 0;
                  }
                &print_perl_obj( $val, $new_pref, $max_items, $file )
              }
            else
              {
                #...print last one no matter what...
                if( $skip && $i < $#{$obj_ref} )
                  {
                    $skip_count++;
                    next;
                  }
                if( $skip_count > 0 )
                  {
                    print {$file} "$pref\[...$skip_count\]\n";
                    $skip_count = 0;
                  }
                &print_perl_obj( \$val, $new_pref, $max_items, $file )
              }
          }
      }
  }
  elsif( ref( $obj_ref ) eq "HASH" )
  {
    @indices = sort keys %{$obj_ref};
    if( $#indices < 0 )
      {
        print {$file} "$pref\{\} undefined\n";
      }
    else
      {
        $skip = 0;
        $skip_count = 0;
        for( $i = 0; $i <= $#indices; $i++ )
          {
            #...determine if should skip printing this one...
            if( $i > $max_items-1 && $max_items >= 0 )
              {
                $skip = 1;
              }
            $new_pref = sprintf( "%s{%s}", $pref, $indices[$i] );
            $val = $$obj_ref{$indices[$i]};
            if( ref( $val ) )
              {
                #...print number skipped...
                if( $skip_count > 0 )
                  {
                    print {$file} "$pref\{...$skip_count\}\n";
                    $skip_count = 0;
                  }
                &print_perl_obj( $val, $new_pref, $max_items, $file )
              }
            else
              {
                if( $skip && $i < $#indices )
                  {
                    $skip_count++;
                    next;
                  }
                if( $skip_count > 0 )
                  {
                    print {$file} "$pref\{...$skip_count\}\n";
                    $skip_count = 0;
                  }
                &print_perl_obj( \$val, $new_pref, $max_items, $file )
              }
          }
      }
  }
  else
  {
     if( defined( $$obj_ref ) )
     {
       print {$file} "$pref = [$$obj_ref]\n";
     }
     else
     {
       print {$file} "$pref undefined\n";
     }
  }
}
#......................................................................
#...Name
#...====
#... log_lanl_usage
#...
#...Purpose
#...=======
#... Log software usage by calling log script NonLicMgr, located in
#... logging directory.
#...
#...Arguments
#...=========
#... $prod             Intent: in
#...                   Perl type: scalar
#...                   the product name
#...
#... $numpes           Intent: in
#...                   Perl type: scalar
#...                   number of processes
#...
#...Program Flow
#...============
#... 1) assign variables
#... 2) system usage program
#...
#.....................................................................
sub log_lanl_usage
  {
    my(
       $prod,
       $numpes,
      ) = @_;
    my(
       $loglocation, # where log file will go
       $logprog, # program to system
       $sys_string,  # string containing command to execute
      );
    #.......................................................
    #...Check for environment variable overriding logpath...
    #.......................................................
    if( defined($ENV{LANL_USAGELOG_PATH}) )
      {
        $loglocation = $ENV{LANL_USAGELOG_PATH};
      }
    else
      {
        #$loglocation = "/usr/projects/packages/lanl_usage_logs";
        $loglocation = "/usr/projects/ups/LANL_USAGE";
      }
    $logprog = "$loglocation/NonLicMgr";
    #.....................................
    #...If log program found execute it...
    #.....................................
    if( -x $logprog )
      {
        $sys_string =
          "(LANL_USAGELOG_PATH=$loglocation; export LANL_USAGELOG_PATH; ".
          "$logprog LANL $prod $numpes ) > /dev/null 2>&1";
        system( $sys_string );
      }
  }
